"""
database.basic is not a real provider but it contains common components for providers that share a common core
and rough strategy:

e.g. mysql and postgres share a ansi sql subprocess-based mysql/psql/mysqldump/pgdump interface

"""